/*
 *  Tcpjunk - TCP protocols testing tool
 *  Contact: oferkv@gmail.com
 *
 *  Tcpjunk is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2, or (at your option) any
 *  later version.
 *
 *  Tcpjunk is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>
#include <ctype.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gtksourceview/gtksourceview.h>
#include <errno.h>
#include <arpa/inet.h>
#include "tcpjunk.h"
#include "client.h"
#include "server.h"
#include "gui.h"
#include "utils.h"
#include "net.h"

struct gui_elements gui_data;

static GtkWidget *get_tab_textview()
  {
  gint current_tab_num = 0;
  GtkWidget *curr_frame, *curr_scrwin, *curr_textview;
  current_tab_num = gtk_notebook_get_current_page(GTK_NOTEBOOK(
	  gui_data.ses_tabs));
  curr_frame = gtk_notebook_get_nth_page(GTK_NOTEBOOK(gui_data.ses_tabs),
	  current_tab_num);
  curr_scrwin = gtk_bin_get_child(GTK_BIN(curr_frame));
  curr_textview = gtk_bin_get_child(GTK_BIN(curr_scrwin));
  return curr_textview;
  }

static gboolean exit_gui(GtkWidget *widget, GdkEvent *event, gpointer data)
  {
  gtk_main_quit();
  return FALSE;
  }

static void change_font(GtkWidget *widget, char *font_string)
  {
  /* textview fonts */
  PangoFontDescription *font_desc;
  font_desc = pango_font_description_from_string(font_string);
  gtk_widget_modify_font(widget, font_desc);
  }

static void gui_update()
  {
  while (gtk_events_pending())
	gtk_main_iteration();
  }

static void gui_clear_session_log()
  {
  GtkTextIter start, end;
  GtkTextBuffer *buffer;
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.session_log_text));
  gtk_text_buffer_get_bounds(buffer, &start, &end);
  gtk_text_buffer_delete(buffer, &start, &end);
  gui_update();
  }

static void gui_clear_status()
  {
  GtkTextIter start, end;
  GtkTextBuffer *buffer;
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.status_text));
  gtk_text_buffer_get_bounds(buffer, &start, &end);
  gtk_text_buffer_delete(buffer, &start, &end);
  gui_update();
  }

static void gui_about(struct tj_data *gui_inst_data)
  {
  gtk_show_about_dialog(NULL, "program-name", "Tcpjunk", "version", VER,
	  "copyright", "Copyright \xC2\xA9 2010 Ofer Kashayov", "comments",
	  "TCP protocols testing tool", "logo_icon_name", "tcpjunk", "website",
	  "http://tcpjunk.googlecode.com", NULL);
  }

gboolean tabs_popup(GtkWidget* widget, GdkEventButton * event)
  {
  if (event->type == GDK_BUTTON_PRESS)
	{
	if (event->button == 3)
	  {
	  gtk_menu_popup(GTK_MENU (gui_data.file_menu), NULL, NULL, NULL, NULL, 1,
		  gtk_get_current_event_time());
	  return TRUE;
	  }
	}
  return FALSE;
  }

void gui_insert_tag(char *tag_text)
  {
  GtkTextBuffer *ses_buf;
  GtkTextIter start, end;

  /* get existing text */
  ses_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(get_tab_textview()));
  gtk_text_buffer_get_bounds(ses_buf, &start, &end);

  /* insert text */
  gtk_text_buffer_insert_at_cursor(ses_buf, tag_text, strlen(tag_text));

  }

void gui_insert_sect()
  {
  char *tag_text = "<sect>user root\n</sect>";
  gui_insert_tag(tag_text);
  }

void gui_insert_rep()
  {
  char *tag_text = "<rep 1 10>text</rep>";
  gui_insert_tag(tag_text);
  }

void gui_insert_sys()
  {
  char *tag_text = "<sys>pwd</sys>";
  gui_insert_tag(tag_text);
  }

void gui_insert_fuzz()
  {
  char *tag_text = "<fuzz alnum 1 10>";
  gui_insert_tag(tag_text);
  }

void gui_insert_char()
  {
  char *tag_text = "<char 5a 10>";
  gui_insert_tag(tag_text);
  }

void gui_insert_enc()
  {
  char *tag_text = "<enc % hex none 2>Hi there!!</enc>";
  gui_insert_tag(tag_text);
  }

void gui_insert_count()
  {
  char *tag_text = "<count hex 1 10>";
  gui_insert_tag(tag_text);
  }

void gui_insert_charr()
  {
  char *tag_text = "<charr>41,42,43</charr>";
  gui_insert_tag(tag_text);
  }

void gui_insert_tamp()
  {
  char *tag_text = "<tamp random 5 15>this is going to be messy</tamp>";
  gui_insert_tag(tag_text);
  }


static void gui_scroll_session_log()
  {

  /* auto scroll */
  GtkTextBuffer *buffer;
  GtkTextIter end;

  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.session_log_text));
  gtk_text_buffer_get_end_iter(buffer, &end);
  gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(gui_data.session_log_text), &end,
	  0, TRUE, 0, 0);
  }

static void gui_show_selection(GtkWidget *text_view,
	struct tj_data *gui_inst_data)
  {
  GtkTextIter start, end;
  GtkTextBuffer *buffer;
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(text_view));
  gtk_text_buffer_get_selection_bounds(buffer, &start, &end);
  int chars_total;
  chars_total = gtk_text_iter_get_offset(&end) - gtk_text_iter_get_offset(
	  &start);
  sprintf(log_msg, "%d characters copied (0x%x)", chars_total, chars_total);
  tj_log(gui_inst_data, log_msg, c_black);
  }

static void text_buffer_changed(struct tj_data *gui_inst_data,
	GtkTextBuffer *sess_textbuf)
  {
  gui_inst_data->sess_data_modified = 1;
  }

static void new_tab(struct tj_data *gui_inst_data)
  {
  GtkWidget *new_lab, *new_textview, *new_frame, *new_scrwin;
  GtkTextBuffer *new_text_buf;

  /* textview */
  static int new_file_count = 1;
  new_textview = gtk_source_view_new();
  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(new_textview), GTK_WRAP_CHAR);
  gtk_source_view_set_highlight_current_line(GTK_SOURCE_VIEW (new_textview),
	  TRUE);
  g_signal_connect(GTK_TEXT_VIEW(new_textview), "copy-clipboard", G_CALLBACK(
		  gui_show_selection), gui_inst_data);

  /* textview buf */
  new_text_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(new_textview));
  gtk_text_buffer_create_tag(new_text_buf, "yellow_hl", "background",
	  "#afffaf", NULL);
  g_signal_connect_swapped(G_OBJECT(new_text_buf), "changed", G_CALLBACK(
		  text_buffer_changed), gui_inst_data);

  /* textview font */
  change_font(new_textview, gui_data.font_string);

  /* scroll win */
  new_scrwin = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(new_scrwin),
	  GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
  gtk_container_add(GTK_CONTAINER(new_scrwin), new_textview);

  /* frame */
  new_frame = gtk_frame_new(NULL);
  gtk_container_add(GTK_CONTAINER(new_frame), new_scrwin);

  /* new tab label */
  new_lab = gtk_label_new(NULL);
  gchar new_title[16];
  sprintf(new_title, "untitled%d", new_file_count++);
  gtk_label_set_text(GTK_LABEL(new_lab), new_title);

  /* create tab */
  gtk_notebook_append_page(GTK_NOTEBOOK(gui_data.ses_tabs), new_frame, new_lab);
  gtk_widget_show_all(gui_data.ses_tabs);
  gint npages = gtk_notebook_get_n_pages(GTK_NOTEBOOK(gui_data.ses_tabs));
  gtk_notebook_set_current_page(GTK_NOTEBOOK(gui_data.ses_tabs), npages - 1);
  gtk_notebook_set_tab_reorderable(GTK_NOTEBOOK(gui_data.ses_tabs), new_frame,
	  TRUE);
  }

static void tab_undo()
  {
  GtkTextBuffer *new_text_buf;
  new_text_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(get_tab_textview()));
  if (gtk_source_buffer_can_undo(GTK_SOURCE_BUFFER(new_text_buf)))
	gtk_source_buffer_undo(GTK_SOURCE_BUFFER(new_text_buf));
  }

static void tab_redo()
  {
  GtkTextBuffer *new_text_buf;
  new_text_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(get_tab_textview()));
  if (gtk_source_buffer_can_redo(GTK_SOURCE_BUFFER(new_text_buf)))
	gtk_source_buffer_redo(GTK_SOURCE_BUFFER(new_text_buf));
  }

static void close_tab(struct tj_data *gui_inst_data)
  {
  gint current_tab_num = 0;
  current_tab_num = gtk_notebook_get_current_page(GTK_NOTEBOOK(
	  gui_data.ses_tabs));
  gint npages = gtk_notebook_get_n_pages(GTK_NOTEBOOK(gui_data.ses_tabs));
  if (npages > 1)
	gtk_notebook_remove_page(GTK_NOTEBOOK(gui_data.ses_tabs), current_tab_num);
  }

static void ipv4_togged(struct tj_data *gui_inst_data)
  {
  if ((gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(gui_data.ipv4_tog)))
	  == TRUE)
	{
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui_data.ipv6_tog), FALSE);
	gui_inst_data->ipv6 = 0;
	}
  else
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui_data.ipv6_tog), TRUE);
  }

static void ipv6_togged(struct tj_data *gui_inst_data)
  {
  if ((gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(gui_data.ipv6_tog)))
	  == TRUE)
	{
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui_data.ipv4_tog), FALSE);
	gui_inst_data->ipv6 = 1;
	}
  else
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui_data.ipv4_tog), TRUE);
  }

static void syntax_highlight(struct tj_data *gui_inst_data)
  {
  GtkTextBuffer *ses_txt_buf, *ses_log_buf;
  GtkTextIter st_start, st_end, sl_start, sl_end;
  GtkTextTag *ses_text_tag, *ses_log_tag;
  GtkTextIter mstart, mend;
  GtkTextTagTable *tag_table;
  gchar *hl_text;
  int matches = 0;

  hl_text = (gchar *) gtk_entry_get_text(GTK_ENTRY(gui_data.syntax_entry));

  /* clear existing tags */
  ses_txt_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(get_tab_textview()));
  gtk_text_buffer_get_bounds(ses_txt_buf, &st_start, &st_end);
  gtk_text_buffer_remove_tag_by_name(ses_txt_buf, "yellow_hl", &st_start,
	  &st_end);
  ses_log_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(
	  gui_data.session_log_text));
  gtk_text_buffer_get_bounds(ses_log_buf, &sl_start, &sl_end);
  gtk_text_buffer_remove_tag_by_name(ses_log_buf, "yellow_hl", &sl_start,
	  &sl_end);

  if (strlen(hl_text) < 1)
	return;

  tag_table = gtk_text_buffer_get_tag_table(ses_txt_buf);
  ses_text_tag = gtk_text_tag_table_lookup(tag_table, "yellow_hl");
  tag_table = gtk_text_buffer_get_tag_table(ses_log_buf);
  ses_log_tag = gtk_text_tag_table_lookup(tag_table, "yellow_hl");

  /* session text */
  while (gtk_text_iter_forward_search(&st_start, hl_text, 0, &mstart, &mend,
	  NULL))
	{
	matches++;
	gtk_text_buffer_apply_tag(ses_txt_buf, ses_text_tag, &mstart, &mend);
	st_start = mend;
	}
  if (matches)
	{
	sprintf(log_msg, "%d highlights in session text", matches);
	tj_log(gui_inst_data, log_msg, c_black);
	matches = 0;
	}

  /* session log */
  while (gtk_text_iter_forward_search(&sl_start, hl_text, 0, &mstart, &mend,
	  NULL))
	{
	matches++;
	gtk_text_buffer_apply_tag(ses_log_buf, ses_log_tag, &mstart, &mend);
	sl_start = mend;
	}
  if (matches)
	{
	sprintf(log_msg, "%d highlights in session log", matches);
	tj_log(gui_inst_data, log_msg, c_black);
	}

  gui_update();
  }

static void gui_connect(struct tj_data *gui_inst_data)
  {
  if ((gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(gui_data.connect_butt)))
	  == FALSE)
	return;

  char *gui_session_text = NULL;

  /* misc GUI cleanup */
  gtk_widget_set_sensitive(GTK_WIDGET(gui_data.connect_butt), FALSE);
  gtk_widget_set_sensitive(GTK_WIDGET(gui_data.accept_butt), FALSE);
  gui_clear_session_log();
  if (gui_data.auto_clearstat == 1)
	gui_clear_status();
  else
	tj_log(gui_inst_data, "-", c_black);

  /* check address */
  char *srv_addr_str;
  srv_addr_str = (gchar *) gtk_entry_get_text(GTK_ENTRY(gui_data.host_entry));
  char *srv_port;
  srv_port = (gchar *) gtk_entry_get_text(GTK_ENTRY(gui_data.port_entry));

  if (gui_inst_data->remote_addr)
	freeaddrinfo(gui_inst_data->remote_addr);

  gui_inst_data->remote_addr = check_address(gui_inst_data, srv_addr_str,
	  srv_port, ai_none);
  if (gui_inst_data->remote_addr == NULL)
	goto done;

  /* timeouts */
  gui_inst_data->conn_timeout = gtk_spin_button_get_value_as_int(
	  GTK_SPIN_BUTTON(gui_data.conn_to_spin));
  gui_inst_data->reply_timeout = gtk_spin_button_get_value_as_int(
	  GTK_SPIN_BUTTON(gui_data.sess_to_spin)) * 1000;

  /* repeat */
  gui_inst_data->keepalive = gtk_spin_button_get_value_as_int(
	  GTK_SPIN_BUTTON(gui_data.repeat_spin));

  /* source port */
  strcpy(gui_inst_data->src_port, gtk_entry_get_text(GTK_ENTRY(
	  gui_data.sport_entry)));

  /* source IP and interface */
  strcpy(gui_inst_data->src_ip, gtk_entry_get_text(
	  GTK_ENTRY(gui_data.sip_entry)));
  strcpy(gui_inst_data->interface, gtk_entry_get_text(
	  GTK_ENTRY(gui_data.ipint_entry)));

  /* send size */
  gui_inst_data->send_size = gtk_spin_button_get_value_as_int(
	  GTK_SPIN_BUTTON(gui_data.send_size_spin));

  /* send delay */
  gui_inst_data->send_delay = gtk_spin_button_get_value_as_int(
	  GTK_SPIN_BUTTON(gui_data.send_delay_spin)) * 1000;

  /* segment size */
  gui_inst_data->max_seg = gtk_spin_button_get_value_as_int(
	  GTK_SPIN_BUTTON(gui_data.seg_size_spin));

  /* load session data from GUI */
  GtkTextIter start, end;
  GtkTextBuffer *buffer;
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(get_tab_textview()));
  gtk_text_buffer_get_bounds(buffer, &start, &end);
  gui_session_text = gtk_text_buffer_get_text(buffer, &start, &end, FALSE);

  /* check session data */
  gui_inst_data->sess_data_len = strlen(gui_session_text);
  if (!gui_inst_data->sess_data_len)
	{
	tj_log(gui_inst_data, "Session data is empty", c_red);
	goto done;
	}
  gui_inst_data->sess_data = tj_realloc(gui_inst_data->sess_data,
	  gui_inst_data->sess_data_len);
  memcpy(gui_inst_data->sess_data, gui_session_text,
	  gui_inst_data->sess_data_len);
  find_sections(gui_inst_data);

  /* connect */
  client_main(gui_inst_data);

  /* syntax highlight */
  syntax_highlight(gui_inst_data);

  /* finish */
  done:

  /* auto scroll session log */
  if (gui_data.auto_scroll == 1)
	{
	gui_scroll_session_log();
	}

  /* reenable gui */
  gtk_widget_set_sensitive(GTK_WIDGET(gui_data.connect_butt), TRUE);
  gtk_widget_set_sensitive(GTK_WIDGET(gui_data.accept_butt), TRUE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui_data.connect_butt), FALSE);

  if (gui_session_text != NULL)
	g_free(gui_session_text);

  return;

  }

static gboolean gui_check_accept(struct tj_data *gui_inst_data)
  {
  if (gui_data.listening == 0)
	return FALSE;

  /* initialize descriptors set */
  int sel_ret;
  fd_set set;
  FD_ZERO (&set);
  FD_SET (gui_data.gui_listen_sock, &set);

  /* initialize timeout */
  struct timeval timeout;
  timeout.tv_sec = 0;
  timeout.tv_usec = 50;

  sel_ret = select(gui_data.gui_listen_sock + 1, &set, NULL, NULL, &timeout);

  if (sel_ret < 1)
	return TRUE;
  else
	{
	gui_accept(gui_inst_data);
	return FALSE;
	}
  }

static gint syntax_entry_key_press(GtkWidget *widget, GdkEventKey *key,
	struct tj_data *gui_inst_data)
  {
  guint keyval;
  keyval = key->keyval;

  switch (keyval)
	{
  case GDK_Return:
  case GDK_KP_Enter:
  case GDK_ISO_Enter:
  case GDK_3270_Enter:
	syntax_highlight(gui_inst_data);

  default:
	return 0;
	}
  }

static void listen_gui_enable(gboolean state)
  {
  gtk_widget_set_sensitive(GTK_WIDGET(gui_data.connect_butt), state);
  gtk_widget_set_sensitive(GTK_WIDGET(gui_data.open_mi), state);
  gtk_entry_set_editable(GTK_ENTRY(gui_data.host_entry), state);
  gtk_entry_set_editable(GTK_ENTRY(gui_data.port_entry), state);
  gtk_entry_set_editable(GTK_ENTRY(gui_data.conn_to_spin), state);
  gtk_entry_set_editable(GTK_ENTRY(gui_data.sess_to_spin), state);
  gtk_text_view_set_editable(GTK_TEXT_VIEW(get_tab_textview()), state);
  gtk_widget_set_sensitive(GTK_WIDGET(gui_data.ipv4_tog), state);
  gtk_widget_set_sensitive(GTK_WIDGET(gui_data.ipv6_tog), state);
  }

static void gui_stop_listening(struct tj_data *gui_inst_data)
  {
  gui_data.listening = 0;
  listen_gui_enable(TRUE);

  tj_log(gui_inst_data, "Aborting...", c_black);
  close(gui_data.gui_listen_sock);
  if (gui_inst_data->local_addr)
	freeaddrinfo(gui_inst_data->local_addr);

  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui_data.accept_butt), FALSE);

  return;
  }

void gui_accept(struct tj_data *gui_inst_data)
  {
  /* accept */
  gui_clear_session_log();
  if (gui_data.auto_clearstat == 1)
	gui_clear_status();
  struct sockaddr_storage cl_addr;
  socklen_t cl_addr_l;
  cl_addr_l = sizeof(cl_addr);

  gui_inst_data->srv_sock = accept(gui_data.gui_listen_sock,
	  (struct sockaddr *) &cl_addr, &cl_addr_l);
  if (gui_inst_data->srv_sock < 0)
	{
	sprintf(log_msg, "failed accept on socket: %s", strerror(errno));
	tj_log(gui_inst_data, log_msg, c_red);
	gui_stop_listening(gui_inst_data);
	return;
	}

  /* start transfers */
  sprintf(log_msg, "TCP connection from %s", addr_ntop(
	  (struct sockaddr *) &cl_addr, cl_addr_l));
  tj_log(gui_inst_data, log_msg, c_black);

  /* stop checking */
  g_timeout_add(500, (GSourceFunc) gui_check_accept, gui_inst_data);

  /* start session */
  server_main(gui_inst_data);

  /* syntax highlight */
  syntax_highlight(gui_inst_data);

  if (gui_data.auto_scroll == 1)
	{
	gui_scroll_session_log();
	}
  }

static void gui_listen(struct tj_data *gui_inst_data)
  {
  if (gui_data.listening == 1)
	{
	gui_stop_listening(gui_inst_data);
	return;
	}

  /* disable GUI */
  listen_gui_enable(FALSE);

  /* clear GUI */
  gui_clear_session_log();
  if (gui_data.auto_clearstat == 1)
	gui_clear_status();

  /* check port */
  const gchar *port_text = gtk_entry_get_text(GTK_ENTRY(gui_data.port_entry));

  if (strlen(port_text) == 0)
	{
	tj_log(gui_inst_data, "Invalid port", c_red);
	goto failure;
	}
  else
	strcpy(gui_inst_data->port, port_text);

  /* timeouts */
  gui_inst_data->conn_timeout = gtk_spin_button_get_value_as_int(
	  GTK_SPIN_BUTTON(gui_data.conn_to_spin));
  gui_inst_data->reply_timeout = gtk_spin_button_get_value_as_int(
	  GTK_SPIN_BUTTON(gui_data.sess_to_spin)) * 1000;

  /* keep alive */
  gui_inst_data->keepalive = gtk_spin_button_get_value_as_int(
	  GTK_SPIN_BUTTON(gui_data.repeat_spin));

  /* send size */
  gui_inst_data->send_size = gtk_spin_button_get_value_as_int(
	  GTK_SPIN_BUTTON(gui_data.send_size_spin));

  /* send delay */
  gui_inst_data->send_delay = gtk_spin_button_get_value_as_int(
	  GTK_SPIN_BUTTON(gui_data.send_delay_spin)) * 1000;

  /* segment size */
  gui_inst_data->max_seg = gtk_spin_button_get_value_as_int(
	  GTK_SPIN_BUTTON(gui_data.seg_size_spin));

  /* get session data from GUI*/
  GtkTextIter start, end;
  GtkTextBuffer *buffer;
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(get_tab_textview()));
  gtk_text_buffer_get_bounds(buffer, &start, &end);

  char *gui_session_text =
	  gtk_text_buffer_get_text(buffer, &start, &end, FALSE);

  /* check session data */
  gui_inst_data->sess_data_len = strlen(gui_session_text);
  if (!gui_inst_data->sess_data_len)
	{
	tj_log(gui_inst_data, "Session data is empty", c_red);
	goto failure;
	}
  gui_inst_data->sess_data = tj_realloc(gui_inst_data->sess_data,
	  gui_inst_data->sess_data_len);
  memcpy(gui_inst_data->sess_data, gui_session_text,
	  gui_inst_data->sess_data_len);
  find_sections(gui_inst_data);

  /* preparation */
  int yes = 1;
  int rslt;
  char bind_ip[128];
  struct sockaddr_storage cl_addr;
  socklen_t cl_addr_l;
  cl_addr_l = sizeof(cl_addr);

  /* set server bind address */
  if (gui_inst_data->ipv6)
	strcpy(bind_ip, "0::0");
  else
	strcpy(bind_ip, "0.0.0.0");
  gui_inst_data->local_addr = check_address(gui_inst_data, bind_ip,
	  gui_inst_data->port, ai_passive);

  /* allocate listening socket */
  gui_data.gui_listen_sock = socket(gui_inst_data->local_addr->ai_family,
	  SOCK_STREAM, 0);
  if (gui_data.gui_listen_sock < 0)
	{
	sprintf(log_msg, "failed to allocate socket: %s", strerror(errno));
	tj_log(gui_inst_data, log_msg, c_red);
	goto failure;
	}

  /* set SO_REUSEADDR on socket */
  rslt = setsockopt(gui_data.gui_listen_sock, SOL_SOCKET, SO_REUSEADDR, &yes,
	  sizeof(int));
  if (rslt < 0)
	{
	sprintf(log_msg, "error setting socket options: %s", strerror(errno));
	tj_log(gui_inst_data, log_msg, c_red);
	goto failure;
	}

  /* bind */
  rslt = bind(gui_data.gui_listen_sock, gui_inst_data->local_addr->ai_addr,
	  gui_inst_data->local_addr->ai_addrlen);
  if (rslt < 0)
	{
	sprintf(log_msg, "failed to bind socket: %s", strerror(errno));
	tj_log(gui_inst_data, log_msg, c_red);
	goto failure;
	}

  /* listen */
  rslt = listen(gui_data.gui_listen_sock, BACKLOG);
  if (rslt < 0)
	{
	sprintf(log_msg, "failed listen on socket: %s", strerror(errno));
	tj_log(gui_inst_data, log_msg, c_red);
	goto failure;
	}

  /* accept */
  g_timeout_add(250, (GSourceFunc) gui_check_accept, gui_inst_data);
  gui_data.listening = 1;
  tj_log(gui_inst_data, "Accepting connections...", c_black);
  return;

  /* something wrong */
  failure:

  gui_stop_listening(gui_inst_data);
  return;

  }

void gui_show_traffic(char *buf, int buf_len, int color,
	struct tj_data *gui_inst_data)
  {
  GtkTextBuffer *ses_log_buf;
  GtkTextIter start, end;
  GtkTextTag *fg_tag;
  GtkTextTagTable *tag_table;
  int i;

  /* disable non printable chars */
  if (!gui_inst_data->hexdump)
	for (i = 0; i < buf_len; i++)
	  {
	  if (isprint(buf[i]) < 1 && isspace(buf[i]) < 1)
		buf[i] = '.';
	  }

  /* get existing text */
  ses_log_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(
	  gui_data.session_log_text));
  gtk_text_buffer_get_bounds(ses_log_buf, &start, &end);
  tag_table = gtk_text_buffer_get_tag_table(ses_log_buf);

  /* choose a color */
  if (color == 0)
	fg_tag = gtk_text_tag_table_lookup(tag_table, "red_fg");
  else
	fg_tag = gtk_text_tag_table_lookup(tag_table, "blue_fg");

  /* insert text */
  if (!gui_inst_data->hexdump)
	gtk_text_buffer_insert_with_tags(ses_log_buf, &end, buf, buf_len, fg_tag,
		NULL);
  else
	{
	char *hex_data_buf = hex_dump(buf, buf_len);
	gtk_text_buffer_insert_with_tags(ses_log_buf, &end, hex_data_buf, strlen(
		hex_data_buf), fg_tag, NULL);
	}

  }

void gui_add_log(const char *message, int color)
  {
  GtkTextBuffer *stat_buf;
  GtkTextIter end;
  GtkTextTag *fg_tag = NULL;
  GtkTextTagTable *tag_table;

  stat_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.status_text));
  gtk_text_buffer_get_end_iter(stat_buf, &end);
  tag_table = gtk_text_buffer_get_tag_table(stat_buf);

  /* choose a color */
  if (color == 0)
	gtk_text_buffer_insert(stat_buf, &end, message, -1);
  else
	{
	if (color == 1)
	  fg_tag = gtk_text_tag_table_lookup(tag_table, "red_fg");
	else if (color == 2)
	  fg_tag = gtk_text_tag_table_lookup(tag_table, "red_fgs");
	else if (color == 3)
	  fg_tag = gtk_text_tag_table_lookup(tag_table, "blue_fgr");
	gtk_text_buffer_insert_with_tags(stat_buf, &end, message, -1, fg_tag, NULL);
	}

  gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(gui_data.status_text), &end, 0,
	  TRUE, 0, 0);

  gui_update();
  }

static void set_tab_name(char *text)
  {
  gint current_tab_num = 0;
  GtkWidget *current_lab, *current_tab;
  current_tab_num = gtk_notebook_get_current_page(GTK_NOTEBOOK(
	  gui_data.ses_tabs));
  current_tab = gtk_notebook_get_nth_page(GTK_NOTEBOOK(gui_data.ses_tabs),
	  current_tab_num);
  current_lab = gtk_notebook_get_tab_label(GTK_NOTEBOOK(gui_data.ses_tabs),
	  current_tab);
  gtk_label_set_text(GTK_LABEL(current_lab), text);
  }

static const gchar *get_tab_name()
  {
  gint current_tab_num = 0;
  GtkWidget *current_lab, *current_tab;
  current_tab_num = gtk_notebook_get_current_page(GTK_NOTEBOOK(
	  gui_data.ses_tabs));
  current_tab = gtk_notebook_get_nth_page(GTK_NOTEBOOK(gui_data.ses_tabs),
	  current_tab_num);
  current_lab = gtk_notebook_get_tab_label(GTK_NOTEBOOK(gui_data.ses_tabs),
	  current_tab);
  const gchar *text;
  text = gtk_label_get_text(GTK_LABEL(current_lab));
  return text;
  }

static void gui_file_open(struct tj_data *gui_inst_data)
  {
  /* open file dialog */
  GtkWidget *dialog;
  dialog = gtk_file_chooser_dialog_new("Open File", GTK_WINDOW(
	  gui_data.window), GTK_FILE_CHOOSER_ACTION_OPEN, GTK_STOCK_CANCEL,
	  GTK_RESPONSE_CANCEL, GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);
  gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(dialog),
	  gui_data.cur_folder);

  if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT)
	{
	char *file_name;
	file_name = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));

	/* load config file */
	strcpy(gui_inst_data->sess_file_name, file_name);
	int rslt = 0;
	rslt = load_session_data(gui_inst_data);
	if (rslt < 0)
	  goto cleanup;

	/* create a new tab */
	new_tab(gui_inst_data);

	/* disable unprintable chars */
	int i;
	int binary_found = 0;
	for (i = 0; i < gui_inst_data->sess_data_len; i++)
	  {
	  if (isprint(gui_inst_data->sess_data[i]) < 1
		  && isspace(gui_inst_data->sess_data[i]) < 1)
		{
		gui_inst_data->sess_data[i] = '.';
		binary_found = 1;
		}
	  }

	if (binary_found)
	  {
	  tj_log(gui_inst_data, "Warning: unprintable characters replaced with .",
		  c_red);
	  tj_log(gui_inst_data,
		  "To use files that contain binary data use tcpjunk in CLI mode",
		  c_red);
	  }

	/* load file in text box */
	GtkTextBuffer *buffer;
	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(get_tab_textview()));
	gtk_text_buffer_set_text(buffer, gui_inst_data->sess_data,
		gui_inst_data->sess_data_len);

	/* display filename in tab */
	char *fn_only;
	for (fn_only = file_name + strlen(file_name); fn_only >= file_name
		&& *fn_only != '/'; fn_only--)
	  ;
	fn_only++;
	set_tab_name(fn_only);

	/* cleanup */
	cleanup:

	g_free(file_name);
	free(gui_data.cur_folder);
	gui_data.cur_folder = gtk_file_chooser_get_current_folder(
		GTK_FILE_CHOOSER(dialog));
	}
  gtk_widget_destroy(dialog);
  }

static void gui_file_save_as(struct tj_data *gui_inst_data)
  {
  /* save dialog */
  GtkWidget *dialog;
  dialog = gtk_file_chooser_dialog_new("Save File", GTK_WINDOW(
	  gui_data.window), GTK_FILE_CHOOSER_ACTION_SAVE, GTK_STOCK_CANCEL,
	  GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);
  gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(dialog),
	  gui_data.cur_folder);
  gtk_file_chooser_set_do_overwrite_confirmation(GTK_FILE_CHOOSER(dialog), TRUE);

  /* click OK */
  if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT)
	{
	char *file_name;
	file_name = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
	GtkTextIter start, end;
	GtkTextBuffer *buffer;
	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(get_tab_textview()));
	gtk_text_buffer_get_bounds(buffer, &start, &end);
	char *gui_session_text = gtk_text_buffer_get_text(buffer, &start, &end,
		FALSE);

	/* open file */
	FILE *save_file;
	if ((save_file = fopen(file_name, "wb")) == NULL)
	  {
	  sprintf(log_msg, "failed to save session file: %s", strerror(errno));
	  tj_log(gui_inst_data, log_msg, c_red);
	  goto cleanup;
	  }
	else
	  {
	  if (strlen(gui_session_text) > 0)
		{
		int i;
		for (i = 0; i <= strlen(gui_session_text) - 1; i++)
		  putc(gui_session_text[i], save_file);
		}
	  fclose(save_file);
	  }
	sprintf(log_msg, "file saved as: %s", file_name);
	tj_log(gui_inst_data, log_msg, c_black);

	/* rename tab */
	char *fn_only;
	for (fn_only = file_name + strlen(file_name); fn_only >= file_name
		&& *fn_only != '/'; fn_only--)
	  ;
	fn_only++;
	set_tab_name(fn_only);

	/* cleanup */
	cleanup:

	g_free(file_name);
	free(gui_data.cur_folder);
	gui_data.cur_folder = gtk_file_chooser_get_current_folder(
		GTK_FILE_CHOOSER(dialog));
	}

  /* click Cancel */
  gtk_widget_destroy(dialog);

  }

static void gui_file_save(struct tj_data *gui_inst_data)
  {
  int i;
  FILE *save_file;
  char *file_name = (char *) get_tab_name();
  char full_path[FILENAME_MAX];
  GtkTextIter start, end;
  sprintf(full_path, "%s/%s", gui_data.cur_folder, file_name);

  /* get text from buffer */
  GtkTextBuffer *buffer;
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(get_tab_textview()));
  gtk_text_buffer_get_bounds(buffer, &start, &end);
  char *gui_session_text =
	  gtk_text_buffer_get_text(buffer, &start, &end, FALSE);

  /* save file */
  if ((save_file = fopen(full_path, "wb")) == NULL)
	{
	sprintf(log_msg, "can not save session file %s: %s", full_path, strerror(
		errno));
	tj_log(gui_inst_data, log_msg, c_red);
	}
  else
	{
	if (strlen(gui_session_text) > 0)
	  for (i = 0; i <= strlen(gui_session_text) - 1; i++)
		putc(gui_session_text[i], save_file);
	fclose(save_file);
	sprintf(log_msg, "file saved as: %s", full_path);
	tj_log(gui_inst_data, log_msg, c_black);
	}
  }

static void set_font(struct tj_data *gui_inst_data, GtkFontButton *font_button)
  {
  strcpy(gui_data.font_string, gtk_font_button_get_font_name(font_button));
  change_font(get_tab_textview(), gui_data.font_string);
  change_font(gui_data.session_log_text, gui_data.font_string);
  }

static void set_sslc(struct tj_data *gui_inst_data, GtkFileChooser *sslc_butt)
  {
  strcpy(gui_inst_data->server_cert, gtk_file_chooser_get_filename(sslc_butt));
  }

static void set_sslk(struct tj_data *gui_inst_data, GtkFileChooser *sslk_butt)
  {
  strcpy(gui_inst_data->server_key, gtk_file_chooser_get_filename(sslk_butt));
  }

static void ssl_toggled(struct tj_data *gui_inst_data, GtkToggleButton *ssl_tog)
  {
  if (gtk_toggle_button_get_active(ssl_tog))
	gui_inst_data->use_ssl = 1;
  else
	gui_inst_data->use_ssl = 0;
  }

static void sett_toggled(struct tj_data *gui_inst_data)
  {
  if ((gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(gui_data.sett_tog)))
	  == TRUE)
	gtk_widget_show(gui_data.settings_box);
  else
	gtk_widget_hide(gui_data.settings_box);
  }

static void wait_data_check_toggled(struct tj_data *gui_inst_data,
	GtkToggleButton *wait_data_checkbox)
  {
  if (gtk_toggle_button_get_active(wait_data_checkbox))
	gui_inst_data->wait_conn = 1;
  else
	gui_inst_data->wait_conn = 0;
  }

static void close_last_check_toggled(struct tj_data *gui_inst_data,
	GtkToggleButton *close_last_checkbox)
  {
  if (gtk_toggle_button_get_active(close_last_checkbox))
	gui_inst_data->no_wait_last = 1;
  else
	gui_inst_data->no_wait_last = 0;
  }

static void auto_scroll_toggled(GtkToggleButton *auto_scr_checkbox)
  {
  if (gtk_toggle_button_get_active(auto_scr_checkbox))
	gui_data.auto_scroll = 1;
  else
	gui_data.auto_scroll = 0;
  }

static void auto_clearstat_toggled(GtkToggleButton *auto_clearstat_checkbox)
  {
  if (gtk_toggle_button_get_active(auto_clearstat_checkbox))
	gui_data.auto_clearstat = 1;
  else
	gui_data.auto_clearstat = 0;
  }

static void hexdump_toggled(struct tj_data *gui_inst_data,
	GtkToggleButton *hexdump_checkbox)
  {
  if (gtk_toggle_button_get_active(hexdump_checkbox))
	gui_inst_data->hexdump = 1;
  else
	gui_inst_data->hexdump = 0;
  }

void gui_init(int argc, char *argv[], struct tj_data *gui_inst_data)
  {
  /* GUI init */
  srand((unsigned int) (time(NULL)));
  gtk_init(&argc, &argv);

  gui_inst_data->gui = 1;
  gui_inst_data->verbose = 1;
  gui_inst_data->debug = 1;
  gui_data.auto_scroll = 0;
  gui_data.listening = 0;
  gui_data.auto_clearstat = 1;
  gui_data.cur_folder = getcwd(NULL, 0);
  strcpy(gui_data.font_string, "Fixed 11");

  /* window */
  gtk_window_set_default_icon_from_file("/usr/share/pixmaps/tcpjunk.png", NULL);
  gui_data.window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(gui_data.window), "Tcpjunk");
  gtk_container_set_border_width(GTK_CONTAINER(gui_data.window), 1);
  g_signal_connect(G_OBJECT(gui_data.window), "delete_event", G_CALLBACK(
		  exit_gui), NULL);

  gtk_window_resize(GTK_WINDOW(gui_data.window), 780, 580);

  /* boxes */
  GtkWidget *main_box, *app_hbox, *session_box, *status_box;
  main_box = gtk_vbox_new(FALSE, 0);
  app_hbox = gtk_hbox_new(FALSE, 0);
  session_box = gtk_vbox_new(FALSE, 0);
  gui_data.settings_box = gtk_vbox_new(FALSE, 0);
  status_box = gtk_vbox_new(FALSE, 0);

  /* check boxes */
  GtkWidget *wait_data_checkbox, *close_last_checkbox, *auto_scr_checkbox,
	  *auto_clearstat_checkbox, *hexdump_checkbox, *ssl_checkbox;
  wait_data_checkbox
	  = gtk_check_button_new_with_label("Wait for data at first");
  close_last_checkbox = gtk_check_button_new_with_label(
	  "Don't wait for data at last");
  auto_scr_checkbox
	  = gtk_check_button_new_with_label("Auto scroll session log");
  auto_clearstat_checkbox
	  = gtk_check_button_new_with_label("Auto clear status");
  hexdump_checkbox = gtk_check_button_new_with_label("Hex dump");
  ssl_checkbox = gtk_check_button_new_with_label("Use SSL");

  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(auto_clearstat_checkbox), TRUE);
  g_signal_connect_swapped(G_OBJECT(wait_data_checkbox), "toggled",
	  G_CALLBACK(wait_data_check_toggled), gui_inst_data);
  g_signal_connect_swapped(G_OBJECT(close_last_checkbox), "toggled",
	  G_CALLBACK(close_last_check_toggled), gui_inst_data);
  g_signal_connect(G_OBJECT(auto_scr_checkbox), "toggled", G_CALLBACK(
		  auto_scroll_toggled), NULL);
  g_signal_connect(G_OBJECT(auto_clearstat_checkbox), "toggled", G_CALLBACK(
		  auto_clearstat_toggled), NULL);
  g_signal_connect_swapped(G_OBJECT(hexdump_checkbox), "toggled", G_CALLBACK(
		  hexdump_toggled), gui_inst_data);
  g_signal_connect_swapped(G_OBJECT(ssl_checkbox), "toggled", G_CALLBACK(
		  ssl_toggled), gui_inst_data);

  /* tabs */
  GtkWidget *settings_tabs;
  gui_data.ses_tabs = gtk_notebook_new();
  settings_tabs = gtk_notebook_new();
  g_signal_connect(G_OBJECT(gui_data.ses_tabs), "button_press_event",
	  G_CALLBACK(tabs_popup), NULL);
  gtk_notebook_set_scrollable(GTK_NOTEBOOK(gui_data.ses_tabs), TRUE);

  /* textviews */
  gui_data.session_log_text = gtk_text_view_new();
  gui_data.status_text = gtk_text_view_new();
  gtk_text_view_set_editable(GTK_TEXT_VIEW(gui_data.session_log_text), FALSE);
  gtk_text_view_set_editable(GTK_TEXT_VIEW(gui_data.status_text), FALSE);
  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(gui_data.session_log_text),
	  GTK_WRAP_CHAR);
  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(gui_data.status_text),
	  GTK_WRAP_WORD);
  g_signal_connect(GTK_TEXT_VIEW(gui_data.session_log_text),
	  "copy-clipboard", G_CALLBACK(gui_show_selection), gui_inst_data);
  change_font(gui_data.session_log_text, gui_data.font_string);

  /* text buffers tags */
  GtkTextBuffer *ses_log_buf, *ses_stat_buf;
  ses_log_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(
	  gui_data.session_log_text));
  ses_stat_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.status_text));
  gtk_text_buffer_create_tag(ses_log_buf, "red_fg", "foreground", "#7f0000",
	  "background", "#fbeded", NULL);
  gtk_text_buffer_create_tag(ses_log_buf, "blue_fg", "foreground", "#00007f",
	  "background", "#ededfb", NULL);
  gtk_text_buffer_create_tag(ses_log_buf, "yellow_hl", "background", "#afffaf",
	  NULL);
  gtk_text_buffer_create_tag(ses_stat_buf, "red_fg", "foreground", "#ff0000",
	  NULL);
  gtk_text_buffer_create_tag(ses_stat_buf, "red_fgs", "foreground", "#7f0000",
	  "background", "#fbeded", NULL);
  gtk_text_buffer_create_tag(ses_stat_buf, "blue_fgr", "foreground", "#00007f",
	  "background", "#ededfb", NULL);

  /* text entries and spin buttons */
  gui_data.host_entry = gtk_entry_new();
  gui_data.port_entry = gtk_entry_new();
  gui_data.sport_entry = gtk_entry_new();
  gui_data.conn_to_spin = gtk_spin_button_new_with_range(0, 99, 1);
  gui_data.sess_to_spin = gtk_spin_button_new_with_range(0, 99999, 100);
  gui_data.repeat_spin = gtk_spin_button_new_with_range(1, 999, 1);
  gui_data.syntax_entry = gtk_entry_new();
  gui_data.sip_entry = gtk_entry_new();
  gui_data.ipint_entry = gtk_entry_new();
  gui_data.send_size_spin = gtk_spin_button_new_with_range(0, 99999, 100);
  gui_data.send_delay_spin = gtk_spin_button_new_with_range(0, 99999, 100);
  gui_data.seg_size_spin = gtk_spin_button_new_with_range(0, 99999, 100);
  gtk_entry_set_width_chars(GTK_ENTRY(gui_data.port_entry), 5);
  gtk_entry_set_width_chars(GTK_ENTRY(gui_data.sport_entry), 5);
  gtk_entry_set_width_chars(GTK_ENTRY(gui_data.host_entry), 30);
  gtk_entry_set_width_chars(GTK_ENTRY(gui_data.syntax_entry), 14);
  gtk_entry_set_width_chars(GTK_ENTRY(gui_data.ipint_entry), 5);
  g_signal_connect(G_OBJECT(gui_data.syntax_entry), "key-press-event",
	  G_CALLBACK(syntax_entry_key_press), gui_inst_data);

  /* set some defaults */
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(gui_data.conn_to_spin),
	  gui_inst_data->conn_timeout);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(gui_data.sess_to_spin),
	  gui_inst_data->reply_timeout / 1000);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(gui_data.send_size_spin),
	  gui_inst_data->send_size);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(gui_data.send_delay_spin),
	  gui_inst_data->send_delay);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(gui_data.seg_size_spin),
	  gui_inst_data->max_seg);

  /* labels */
  GtkWidget *host_lab, *port_lab, *sock_to_lab, *init_to_lab, *repeat_lab,
	  *syntax_lab, *sport_lab, *font_lab, *sslc_lab, *sslk_lab, *sip_lab,
	  *ipint_lab, *srs_lab, *srd_lab, *seg_size_lab;
  GtkWidget *conn_lab, *display_lab, *ssl_lab;
  host_lab = gtk_label_new("Host/IP ");
  port_lab = gtk_label_new(" Service/Port ");
  sport_lab = gtk_label_new("Source port");
  init_to_lab = gtk_label_new("Connection timeout (s) ");
  sock_to_lab = gtk_label_new("Session timeout (ms) ");
  repeat_lab = gtk_label_new("Repeat session");
  syntax_lab = gtk_label_new("Highlight ");
  font_lab = gtk_label_new("Session font");
  sslc_lab = gtk_label_new("SSL server certificate");
  sslk_lab = gtk_label_new("SSL server key");
  sip_lab = gtk_label_new("Source IP");
  ipint_lab = gtk_label_new("Source IP interface");
  srs_lab = gtk_label_new("Send buffer size");
  srd_lab = gtk_label_new("Send delay (ms)");
  conn_lab = gtk_label_new("Connection");
  display_lab = gtk_label_new("Display");
  ssl_lab = gtk_label_new("SSL");
  seg_size_lab = gtk_label_new("Segment size");
  gtk_misc_set_alignment(GTK_MISC(sport_lab), 0.0, 1.0);
  gtk_misc_set_alignment(GTK_MISC(init_to_lab), 0.0, 1.0);
  gtk_misc_set_alignment(GTK_MISC(sock_to_lab), 0.0, 1.0);
  gtk_misc_set_alignment(GTK_MISC(repeat_lab), 0.0, 1.0);
  gtk_misc_set_alignment(GTK_MISC(syntax_lab), 0.0, 1.0);
  gtk_misc_set_alignment(GTK_MISC(font_lab), 0.0, 1.0);
  gtk_misc_set_alignment(GTK_MISC(sslc_lab), 0.0, 1.0);
  gtk_misc_set_alignment(GTK_MISC(sslk_lab), 0.0, 1.0);
  gtk_misc_set_alignment(GTK_MISC(sip_lab), 0.0, 1.0);
  gtk_misc_set_alignment(GTK_MISC(ipint_lab), 0.0, 1.0);
  gtk_misc_set_alignment(GTK_MISC(srs_lab), 0.0, 1.0);
  gtk_misc_set_alignment(GTK_MISC(srd_lab), 0.0, 1.0);
  gtk_misc_set_alignment(GTK_MISC(seg_size_lab), 0.0, 1.0);

  /* tables */
  GtkWidget *conf_table, *display_table, *ssl_table;
  conf_table = gtk_table_new(20, 2, FALSE);
  display_table = gtk_table_new(4, 2, FALSE);
  ssl_table = gtk_table_new(4, 2, FALSE);

  /* frames */
  GtkWidget *session_frame, *session_log_frame, *status_frame;
  session_frame = gtk_frame_new(NULL);
  session_log_frame = gtk_frame_new(NULL);
  status_frame = gtk_frame_new(NULL);

  /* menus */
  GtkWidget *edit_menu, *tags_menu, *help_menu;
  GtkWidget *file_mi, *edit_mi, *tags_mi, *help_mi;
  gui_data.file_menu = gtk_menu_new();
  edit_menu = gtk_menu_new();
  tags_menu = gtk_menu_new();
  help_menu = gtk_menu_new();
  file_mi = gtk_menu_item_new_with_label("File");
  edit_mi = gtk_menu_item_new_with_label("Edit");
  tags_mi = gtk_menu_item_new_with_label("Tags");
  help_mi = gtk_menu_item_new_with_label("Help");

  /* file, edit and help menu items */
  GtkWidget *new_mi, *save_mi, *saveas_mi, *quit_mi, *close_mi, *undo_mi,
	  *redo_mi, *about_mi;
  new_mi = gtk_image_menu_item_new_with_label("New tab");
  gui_data.open_mi = gtk_image_menu_item_new_with_label("Open");
  save_mi = gtk_image_menu_item_new_with_label("Save");
  saveas_mi = gtk_menu_item_new_with_label("Save as");
  close_mi = gtk_image_menu_item_new_with_label("Close tab");
  quit_mi = gtk_image_menu_item_new_with_label("Quit");
  undo_mi = gtk_image_menu_item_new_with_label("Undo");
  redo_mi = gtk_image_menu_item_new_with_label("Redo");
  about_mi = gtk_image_menu_item_new_with_label("About");
  gtk_menu_shell_append(GTK_MENU_SHELL(gui_data.file_menu), new_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(gui_data.file_menu), close_mi);
  GtkWidget *sep1 = gtk_separator_menu_item_new();
  gtk_menu_shell_append(GTK_MENU_SHELL(gui_data.file_menu), sep1);
  gtk_menu_shell_append(GTK_MENU_SHELL(gui_data.file_menu), gui_data.open_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(gui_data.file_menu), save_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(gui_data.file_menu), saveas_mi);
  GtkWidget *sep2 = gtk_separator_menu_item_new();
  gtk_menu_shell_append(GTK_MENU_SHELL(gui_data.file_menu), sep2);
  gtk_menu_shell_append(GTK_MENU_SHELL(gui_data.file_menu), quit_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(edit_menu), undo_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(edit_menu), redo_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(help_menu), about_mi);

  /* tags menu items */
  GtkWidget *section_mi, *repeat_mi, *system_mi, *fuzz_mi, *enc_mi, *char_mi,
	  *count_mi, *charr_mi, *tamp_mi;
  section_mi = gtk_menu_item_new_with_label("Section");
  repeat_mi = gtk_menu_item_new_with_label("Repeat");
  system_mi = gtk_menu_item_new_with_label("System");
  fuzz_mi = gtk_menu_item_new_with_label("Fuzz");
  enc_mi = gtk_menu_item_new_with_label("Encode");
  char_mi = gtk_menu_item_new_with_label("Charaters");
  count_mi = gtk_menu_item_new_with_label("Counter");
  charr_mi = gtk_menu_item_new_with_label("Character Array");
  tamp_mi = gtk_menu_item_new_with_label("Tamper");
  gtk_menu_shell_append(GTK_MENU_SHELL(tags_menu), section_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(tags_menu), repeat_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(tags_menu), system_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(tags_menu), fuzz_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(tags_menu), char_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(tags_menu), enc_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(tags_menu), count_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(tags_menu), charr_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(tags_menu), tamp_mi);

  /* menu bar */
  GtkWidget *menu_bar;
  menu_bar = gtk_menu_bar_new();
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), file_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), edit_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), tags_mi);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), help_mi);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(file_mi), gui_data.file_menu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(edit_mi), edit_menu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(tags_mi), tags_menu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(help_mi), help_menu);

  /* menu images */
  GtkWidget *tmp_mi_image;
  tmp_mi_image = gtk_image_new_from_stock(GTK_STOCK_NEW, GTK_ICON_SIZE_MENU);
  gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(new_mi), tmp_mi_image);
  tmp_mi_image = gtk_image_new_from_stock(GTK_STOCK_OPEN, GTK_ICON_SIZE_MENU);
  gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(gui_data.open_mi),
	  tmp_mi_image);
  tmp_mi_image = gtk_image_new_from_stock(GTK_STOCK_SAVE, GTK_ICON_SIZE_MENU);
  gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(save_mi), tmp_mi_image);
  tmp_mi_image = gtk_image_new_from_stock(GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU);
  gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(close_mi), tmp_mi_image);
  tmp_mi_image = gtk_image_new_from_stock(GTK_STOCK_QUIT, GTK_ICON_SIZE_MENU);
  gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(quit_mi), tmp_mi_image);
  tmp_mi_image = gtk_image_new_from_stock(GTK_STOCK_UNDO, GTK_ICON_SIZE_MENU);
  gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(undo_mi), tmp_mi_image);
  tmp_mi_image = gtk_image_new_from_stock(GTK_STOCK_REDO, GTK_ICON_SIZE_MENU);
  gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(redo_mi), tmp_mi_image);
  tmp_mi_image = gtk_image_new_from_stock(GTK_STOCK_ABOUT, GTK_ICON_SIZE_MENU);
  gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(about_mi), tmp_mi_image);

  /* menu signals */
  g_signal_connect_swapped(G_OBJECT(gui_data.open_mi), "activate",
	  G_CALLBACK(gui_file_open), gui_inst_data);
  g_signal_connect_swapped(G_OBJECT(saveas_mi), "activate", G_CALLBACK(
		  gui_file_save_as), gui_inst_data);
  g_signal_connect_swapped(G_OBJECT(save_mi), "activate", G_CALLBACK(
		  gui_file_save), gui_inst_data);
  g_signal_connect_swapped(G_OBJECT(new_mi), "activate", G_CALLBACK(new_tab),
	  gui_inst_data);
  g_signal_connect_swapped(G_OBJECT(close_mi), "activate", G_CALLBACK(
		  close_tab), gui_inst_data);
  g_signal_connect(G_OBJECT(quit_mi), "activate", G_CALLBACK(exit_gui), NULL);
  g_signal_connect_swapped(G_OBJECT(about_mi), "activate", G_CALLBACK(
		  gui_about), gui_inst_data);
  g_signal_connect(G_OBJECT(section_mi), "activate", G_CALLBACK(
		  gui_insert_sect), NULL);
  g_signal_connect(G_OBJECT(repeat_mi), "activate",
	  G_CALLBACK(gui_insert_rep), NULL);
  g_signal_connect(G_OBJECT(system_mi), "activate",
	  G_CALLBACK(gui_insert_sys), NULL);
  g_signal_connect(G_OBJECT(fuzz_mi), "activate",
	  G_CALLBACK(gui_insert_fuzz), NULL);
  g_signal_connect(G_OBJECT(char_mi), "activate",
	  G_CALLBACK(gui_insert_char), NULL);
  g_signal_connect(G_OBJECT(enc_mi), "activate", G_CALLBACK(gui_insert_enc),
	  NULL);
  g_signal_connect(G_OBJECT(count_mi), "activate", G_CALLBACK(
		  gui_insert_count), NULL);
  g_signal_connect(G_OBJECT(charr_mi), "activate", G_CALLBACK(
		  gui_insert_charr), NULL);
  g_signal_connect(G_OBJECT(tamp_mi), "activate", G_CALLBACK(
  		  gui_insert_tamp), NULL);
  g_signal_connect(G_OBJECT(undo_mi), "activate", G_CALLBACK(
		  tab_undo), NULL);
  g_signal_connect(G_OBJECT(redo_mi), "activate", G_CALLBACK(
		  tab_redo), NULL);

  /* scroll windows */
  GtkWidget *ses_set_scr_win, *session_log_scr_win, *status_scr_win;
  ses_set_scr_win = gtk_scrolled_window_new(NULL, NULL);
  session_log_scr_win = gtk_scrolled_window_new(NULL, NULL);
  status_scr_win = gtk_scrolled_window_new(NULL, NULL);
  gtk_container_add(GTK_CONTAINER(session_log_scr_win),
	  gui_data.session_log_text);
  gtk_container_add(GTK_CONTAINER(status_scr_win), gui_data.status_text);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(ses_set_scr_win),
	  GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(session_log_scr_win),
	  GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(status_scr_win),
	  GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);

  /* font button */
  GtkWidget *font_butt;
  font_butt = gtk_font_button_new_with_font(gui_data.font_string);
  g_signal_connect_swapped(G_OBJECT(font_butt), "font-set",
	  G_CALLBACK(set_font), (gpointer) gui_inst_data);

  /* connect button */
  gui_data.connect_butt = gtk_toggle_button_new_with_label(" Connect");
  GTK_WIDGET_UNSET_FLAGS(gui_data.connect_butt, GTK_CAN_FOCUS);
  g_signal_connect_swapped(G_OBJECT(gui_data.connect_butt), "toggled",
	  G_CALLBACK(gui_connect), gui_inst_data);

  /* accept button */
  gui_data.accept_butt = gtk_toggle_button_new_with_label(" Accept");
  GTK_WIDGET_UNSET_FLAGS(gui_data.accept_butt, GTK_CAN_FOCUS);
  g_signal_connect_swapped(G_OBJECT(gui_data.accept_butt), "toggled",
	  G_CALLBACK(gui_listen), (gpointer) gui_inst_data);

  /* ipv4 button */
  gui_data.ipv4_tog = gtk_toggle_button_new_with_label("IPv4");
  GTK_WIDGET_UNSET_FLAGS(gui_data.ipv4_tog, GTK_CAN_FOCUS);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui_data.ipv4_tog), TRUE);
  g_signal_connect_swapped(G_OBJECT(gui_data.ipv4_tog), "toggled",
	  G_CALLBACK(ipv4_togged), gui_inst_data);

  /* ipv6 button */
  gui_data.ipv6_tog = gtk_toggle_button_new_with_label("IPv6");
  GTK_WIDGET_UNSET_FLAGS(gui_data.ipv6_tog, GTK_CAN_FOCUS);
  g_signal_connect_swapped(G_OBJECT(gui_data.ipv6_tog), "toggled",
	  G_CALLBACK(ipv6_togged), gui_inst_data);

  /* settings button */
  gui_data.sett_tog = gtk_toggle_button_new_with_label(" Settings");
  g_signal_connect_swapped(G_OBJECT(gui_data.sett_tog), "toggled", G_CALLBACK(
		  sett_toggled), gui_inst_data);

  /* SSL server certificate button */
  GtkWidget *sslc_butt;
  sslc_butt = gtk_file_chooser_button_new("Select a File",
	  GTK_FILE_CHOOSER_ACTION_OPEN);
  g_signal_connect_swapped(G_OBJECT(sslc_butt), "file-set",
	  G_CALLBACK(set_sslc), (gpointer) gui_inst_data);

  /* SSL server key button */
  GtkWidget *sslk_butt;
  sslk_butt = gtk_file_chooser_button_new("Select a File",
	  GTK_FILE_CHOOSER_ACTION_OPEN);
  g_signal_connect_swapped(G_OBJECT(sslk_butt), "file-set",
	  G_CALLBACK(set_sslk), (gpointer) gui_inst_data);

  /* button images */
  GtkWidget *tmp_image;
  tmp_image = gtk_image_new_from_stock(GTK_STOCK_CONNECT, GTK_ICON_SIZE_MENU);
  gtk_button_set_image(GTK_BUTTON(gui_data.connect_butt), tmp_image);
  gtk_widget_show(tmp_image);
  tmp_image = gtk_image_new_from_stock(GTK_STOCK_NETWORK, GTK_ICON_SIZE_MENU);
  gtk_button_set_image(GTK_BUTTON(gui_data.accept_butt), tmp_image);
  gtk_widget_show(tmp_image);
  tmp_image = gtk_image_new_from_stock(GTK_STOCK_PREFERENCES,
	  GTK_ICON_SIZE_MENU);
  gtk_button_set_image(GTK_BUTTON(gui_data.sett_tog), tmp_image);
  gtk_widget_show(tmp_image);

  /* alignments */
  GtkWidget *port_alig, *conn_to_alig, *sess_to_alig, *conf_table_alig,
	  *ssl_table_alig, *display_table_alig, *host_alig, *repeat_alig,
	  *syntax_alig, *sport_alig;
  port_alig = gtk_alignment_new(0, 0, 0, 0);
  sport_alig = gtk_alignment_new(0, 0, 0, 0);
  host_alig = gtk_alignment_new(0, 0, 0, 0);
  conn_to_alig = gtk_alignment_new(0, 0, 0, 0);
  sess_to_alig = gtk_alignment_new(0, 0, 0, 0);
  repeat_alig = gtk_alignment_new(0, 0, 0, 0);
  conf_table_alig = gtk_alignment_new(0, 0, 1, 0);
  ssl_table_alig = gtk_alignment_new(0, 0, 1, 0);
  display_table_alig = gtk_alignment_new(0, 0, 1, 0);
  syntax_alig = gtk_alignment_new(0, 0, 0, 0);
  gtk_container_add(GTK_CONTAINER(port_alig), gui_data.port_entry);
  gtk_container_add(GTK_CONTAINER(sport_alig), gui_data.sport_entry);
  gtk_container_add(GTK_CONTAINER(host_alig), gui_data.host_entry);
  gtk_container_add(GTK_CONTAINER(conn_to_alig), gui_data.conn_to_spin);
  gtk_container_add(GTK_CONTAINER(sess_to_alig), gui_data.sess_to_spin);
  gtk_container_add(GTK_CONTAINER(conf_table_alig), conf_table);
  gtk_container_add(GTK_CONTAINER(ssl_table_alig), ssl_table);
  gtk_container_add(GTK_CONTAINER(display_table_alig), display_table);
  gtk_container_add(GTK_CONTAINER(repeat_alig), gui_data.repeat_spin);
  gtk_container_add(GTK_CONTAINER(syntax_alig), gui_data.syntax_entry);
  gtk_alignment_set_padding(GTK_ALIGNMENT(conf_table_alig), 3, 3, 5, 5);
  gtk_alignment_set_padding(GTK_ALIGNMENT(display_table_alig), 3, 3, 5, 5);
  gtk_alignment_set_padding(GTK_ALIGNMENT(ssl_table_alig), 3, 3, 5, 5);

  /* vpanes */
  GtkWidget *edit_pan, *session_pan;
  edit_pan = gtk_vpaned_new();
  session_pan = gtk_vpaned_new();
  gtk_paned_set_position(GTK_PANED(session_pan), 200);
  gtk_paned_set_position(GTK_PANED(edit_pan), 200);

  /* pack sessions texts */
  gtk_container_add(GTK_CONTAINER(session_frame), ses_set_scr_win);
  gtk_container_add(GTK_CONTAINER(session_log_frame), session_log_scr_win);
  gtk_container_add(GTK_CONTAINER(status_frame), status_scr_win);
  gtk_box_pack_start(GTK_BOX(status_box), GTK_WIDGET(status_frame), TRUE, TRUE,
	  1);

  /* toolbar */
  GtkWidget *addr_toolbar;
  addr_toolbar = gtk_toolbar_new();

  GtkToolItem *host_alig_ti, *host_lab_ti, *port_lab_ti, *port_alig_ti,
	  *ipv4_tog_ti, *ipv6_tog_ti, *conn_ti, *acc_ti, *sett_ti;
  GtkToolItem *tbsep1, *tbsep2, *tbsep3;

  host_alig_ti = gtk_tool_item_new();
  host_lab_ti = gtk_tool_item_new();
  port_lab_ti = gtk_tool_item_new();
  port_alig_ti = gtk_tool_item_new();
  ipv4_tog_ti = gtk_tool_item_new();
  ipv6_tog_ti = gtk_tool_item_new();
  conn_ti = gtk_tool_item_new();
  acc_ti = gtk_tool_item_new();
  sett_ti = gtk_tool_item_new();
  tbsep1 = gtk_separator_tool_item_new();
  tbsep2 = gtk_separator_tool_item_new();
  tbsep3 = gtk_separator_tool_item_new();

  gtk_container_add(GTK_CONTAINER (host_lab_ti), host_lab);
  gtk_container_add(GTK_CONTAINER (host_alig_ti), host_alig);
  gtk_container_add(GTK_CONTAINER (port_lab_ti), port_lab);
  gtk_container_add(GTK_CONTAINER (port_alig_ti), port_alig);
  gtk_container_add(GTK_CONTAINER (ipv4_tog_ti), gui_data.ipv4_tog);
  gtk_container_add(GTK_CONTAINER (ipv6_tog_ti), gui_data.ipv6_tog);
  gtk_container_add(GTK_CONTAINER (conn_ti), gui_data.connect_butt);
  gtk_container_add(GTK_CONTAINER (acc_ti), gui_data.accept_butt);
  gtk_container_add(GTK_CONTAINER (sett_ti), gui_data.sett_tog);

  gtk_toolbar_insert(GTK_TOOLBAR(addr_toolbar), host_lab_ti, -1);
  gtk_toolbar_insert(GTK_TOOLBAR(addr_toolbar), host_alig_ti, -1);
  gtk_toolbar_insert(GTK_TOOLBAR(addr_toolbar), port_lab_ti, -1);
  gtk_toolbar_insert(GTK_TOOLBAR(addr_toolbar), port_alig_ti, -1);
  gtk_toolbar_insert(GTK_TOOLBAR(addr_toolbar), tbsep1, -1);
  gtk_toolbar_insert(GTK_TOOLBAR(addr_toolbar), ipv4_tog_ti, -1);
  gtk_toolbar_insert(GTK_TOOLBAR(addr_toolbar), ipv6_tog_ti, -1);
  gtk_toolbar_insert(GTK_TOOLBAR(addr_toolbar), tbsep2, -1);
  gtk_toolbar_insert(GTK_TOOLBAR(addr_toolbar), conn_ti, -1);
  gtk_toolbar_insert(GTK_TOOLBAR(addr_toolbar), acc_ti, -1);
  gtk_toolbar_insert(GTK_TOOLBAR(addr_toolbar), tbsep3, -1);
  gtk_toolbar_insert(GTK_TOOLBAR(addr_toolbar), sett_ti, -1);

  /* pack connection table */
  gtk_table_attach(GTK_TABLE(conf_table), wait_data_checkbox, 0, 2, 6, 7,
	  GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), close_last_checkbox, 0, 2, 7, 8,
	  GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), init_to_lab, 0, 1, 8, 9, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), conn_to_alig, 1, 2, 8, 9, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), sock_to_lab, 0, 1, 10, 11, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), sess_to_alig, 1, 2, 10, 11,
	  GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), repeat_lab, 0, 1, 11, 12, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), repeat_alig, 1, 2, 11, 12, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), sport_lab, 0, 1, 12, 13, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 0);
  gtk_table_attach(GTK_TABLE(conf_table), sport_alig, 1, 2, 12, 13, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), sip_lab, 0, 2, 13, 14, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), gui_data.sip_entry, 0, 2, 14, 15,
	  GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), ipint_lab, 0, 1, 15, 16, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), gui_data.ipint_entry, 1, 2, 15, 16,
	  GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), srs_lab, 0, 1, 16, 17, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), gui_data.send_size_spin, 1, 2, 16,
	  17, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), srd_lab, 0, 1, 17, 18, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(conf_table), gui_data.send_delay_spin, 1, 2, 17,
	  18, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 2);


   gtk_table_attach(GTK_TABLE(conf_table), seg_size_lab, 0, 1, 18,
      	  19, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 2);
   gtk_table_attach(GTK_TABLE(conf_table), gui_data.seg_size_spin, 1, 2, 18,
    	  19, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 2);

  /* pack ssl table */
  gtk_table_attach(GTK_TABLE(ssl_table), ssl_checkbox, 0, 2, 0, 1, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(ssl_table), sslc_lab, 0, 2, 1, 2, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(ssl_table), sslc_butt, 0, 2, 2, 3, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(ssl_table), sslk_lab, 0, 2, 3, 4, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(ssl_table), sslk_butt, 0, 2, 4, 5, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);

  /* pack display table */
  gtk_table_attach(GTK_TABLE(display_table), hexdump_checkbox, 0, 2, 0, 1,
	  GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(display_table), auto_scr_checkbox, 0, 2, 1, 2,
	  GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(display_table), auto_clearstat_checkbox, 0, 2, 2,
	  3, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(display_table), syntax_lab, 0, 1, 3, 4, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(display_table), syntax_alig, 1, 2, 3, 4,
	  GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(display_table), font_lab, 0, 1, 4, 5, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(display_table), font_butt, 0, 2, 5, 6, GTK_EXPAND
	  | GTK_FILL, GTK_SHRINK, 0, 2);

  /* pack settings box */
  gtk_notebook_append_page(GTK_NOTEBOOK(settings_tabs), conf_table_alig,
	  conn_lab);
  gtk_notebook_append_page(GTK_NOTEBOOK(settings_tabs), display_table_alig,
	  display_lab);
  gtk_notebook_append_page(GTK_NOTEBOOK(settings_tabs), ssl_table_alig, ssl_lab);
  gtk_box_pack_start(GTK_BOX(gui_data.settings_box), GTK_WIDGET(settings_tabs),
	  TRUE, TRUE, 1);

  /* pack session box */
  gtk_paned_add1(GTK_PANED(session_pan), session_log_frame);
  gtk_paned_add2(GTK_PANED(session_pan), status_box);
  gtk_paned_add1(GTK_PANED(edit_pan), gui_data.ses_tabs);
  gtk_paned_add2(GTK_PANED(edit_pan), session_pan);
  gtk_box_pack_start(GTK_BOX(session_box), GTK_WIDGET(edit_pan), TRUE, TRUE, 1);

  /* pack window */
  gtk_box_pack_start(GTK_BOX(app_hbox), GTK_WIDGET(session_box), TRUE, TRUE, 1);
  gtk_box_pack_start(GTK_BOX(app_hbox), GTK_WIDGET(gui_data.settings_box),
	  FALSE, FALSE, 1);
  gtk_box_pack_start(GTK_BOX(main_box), GTK_WIDGET(menu_bar), FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(main_box), GTK_WIDGET(addr_toolbar), FALSE, FALSE,
	  0);
  gtk_box_pack_start(GTK_BOX(main_box), GTK_WIDGET(app_hbox), TRUE, TRUE, 1);
  gtk_container_add(GTK_CONTAINER(gui_data.window), main_box);
  new_tab(gui_inst_data);
  gtk_widget_show_all(gui_data.window);
  gtk_widget_hide(gui_data.settings_box);

  /* accelerators */
  GtkAccelGroup *ac_grp;
  ac_grp = gtk_accel_group_new();
  gtk_window_add_accel_group(GTK_WINDOW(gui_data.window), ac_grp);
  gtk_widget_add_accelerator(gui_data.connect_butt, "activate", ac_grp, GDK_E,
	  GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
  gtk_widget_add_accelerator(gui_data.accept_butt, "activate", ac_grp, GDK_L,
	  GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
  gtk_widget_add_accelerator(gui_data.open_mi, "activate", ac_grp, GDK_O,
	  GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
  gtk_widget_add_accelerator(close_mi, "activate", ac_grp, GDK_W,
	  GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
  gtk_widget_add_accelerator(new_mi, "activate", ac_grp, GDK_N,
	  GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
  gtk_widget_add_accelerator(save_mi, "activate", ac_grp, GDK_S,
	  GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
  gtk_widget_add_accelerator(undo_mi, "activate", ac_grp, GDK_Z,
	  GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
  gtk_widget_add_accelerator(redo_mi, "activate", ac_grp, GDK_Z,
	  GDK_CONTROL_MASK | GDK_SHIFT_MASK, GTK_ACCEL_VISIBLE);

  /* tooltips */
  gtk_widget_set_tooltip_text(gui_data.connect_butt, "Ctrl+E");
  gtk_widget_set_tooltip_text(gui_data.accept_butt, "Ctrl+L");

  /* show version */
  sprintf(log_msg, "Tcpjunk %s", VER);
  tj_log(gui_inst_data, log_msg, c_black);

  gtk_main();

  }

